import os
import random
import sys
import time
import unittest
from typing import List

sys.path.insert(0, os.path.split(os.path.split(os.path.abspath(__file__))[0])[0])

from src.dtppy import Server, Client, server, client
from src.dtppy.util import encode_message_size, decode_message_size
from src.dtppy.crypto import new_rsa_keys, rsa_encrypt, rsa_decrypt, new_aes_key, aes_encrypt, aes_decrypt

from expect_map import ExpectMap

WAIT_TIME = 0.2


class Custom:
    def __init__(self, a: int, b: str, c: List[str]):
        self.a = a
        self.b = b
        self.c = c

    def __eq__(self, other):
        return self.a == other.a and self.b == other.b and self.c == other.c


class TestUtil(unittest.TestCase):
    def test_encode_message_size(self):
        self.assertEqual(encode_message_size(0), bytes([0, 0, 0, 0, 0]))
        self.assertEqual(encode_message_size(1), bytes([0, 0, 0, 0, 1]))
        self.assertEqual(encode_message_size(255), bytes([0, 0, 0, 0, 255]))
        self.assertEqual(encode_message_size(256), bytes([0, 0, 0, 1, 0]))
        self.assertEqual(encode_message_size(257), bytes([0, 0, 0, 1, 1]))
        self.assertEqual(encode_message_size(4311810305), bytes([1, 1, 1, 1, 1]))
        self.assertEqual(encode_message_size(4328719365), bytes([1, 2, 3, 4, 5]))
        self.assertEqual(encode_message_size(47362409218), bytes([11, 7, 5, 3, 2]))
        self.assertEqual(encode_message_size(1099511627775), bytes([255, 255, 255, 255, 255]))

    def test_decode_message_size(self):
        self.assertEqual(decode_message_size(bytes([0, 0, 0, 0, 0])), 0)
        self.assertEqual(decode_message_size(bytes([0, 0, 0, 0, 1])), 1)
        self.assertEqual(decode_message_size(bytes([0, 0, 0, 0, 255])), 255)
        self.assertEqual(decode_message_size(bytes([0, 0, 0, 1, 0])), 256)
        self.assertEqual(decode_message_size(bytes([0, 0, 0, 1, 1])), 257)
        self.assertEqual(decode_message_size(bytes([1, 1, 1, 1, 1])), 4311810305)
        self.assertEqual(decode_message_size(bytes([1, 2, 3, 4, 5])), 4328719365)
        self.assertEqual(decode_message_size(bytes([11, 7, 5, 3, 2])), 47362409218)
        self.assertEqual(decode_message_size(bytes([255, 255, 255, 255, 255])), 1099511627775)


class TestCrypto(unittest.TestCase):
    def test_rsa(self):
        message = "Hello, RSA!"

        public_key, private_key = new_rsa_keys()
        encrypted = rsa_encrypt(public_key, message.encode("utf-8"))
        decrypted = rsa_decrypt(private_key, encrypted)
        decrypted_message = decrypted.decode("utf-8")

        self.assertEqual(decrypted_message, message)
        self.assertNotEqual(encrypted, message.encode("utf-8"))

    def test_aes(self):
        message = "Hello, AES!"

        key = new_aes_key()
        encrypted = aes_encrypt(key, message.encode("utf-8"))
        decrypted = aes_decrypt(key, encrypted)
        decrypted_message = decrypted.decode("utf-8")

        self.assertEqual(decrypted_message, message)
        self.assertNotEqual(encrypted, message.encode("utf-8"))

    def test_encrypting_aes_key_with_rsa(self):
        public_key, private_key = new_rsa_keys()
        key = new_aes_key()

        encrypted_key = rsa_encrypt(public_key, key)
        decrypted_key = rsa_decrypt(private_key, encrypted_key)

        self.assertEqual(decrypted_key, key)
        self.assertNotEqual(encrypted_key, key)


class TestMain(unittest.TestCase):
    def test_server_serve(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 0,
            "server connect": 0,
            "server disconnect": 0,
        })
        self.assertEqual(expected.get_expected(), {
            "server receive": 0,
            "server connect": 0,
            "server disconnect": 0,
        })
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

        def on_receive(client_id, data):
            expected.received("server receive")

        def on_connect(client_id):
            expected.received("server connect")

        def on_disconnect(client_id):
            expected.received("server disconnect")

        # Create server
        s = Server(on_receive=on_receive, on_connect=on_connect, on_disconnect=on_disconnect)
        self.assertFalse(s.serving())

        # Start server
        s.start()
        self.assertTrue(s.serving())
        time.sleep(WAIT_TIME)

        # Check server address info
        # print(f"Server address: {s.get_addr()}")

        # Stop server
        s.stop()
        self.assertFalse(s.serving())
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.get_expected(), {
            "server receive": 0,
            "server connect": 0,
            "server disconnect": 0,
        })
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_addresses(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 0,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": 0,
            "client disconnected": 0
        })

        def server_on_receive(client_id, data):
            expected.received("server receive")

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        self.assertFalse(s.serving())
        s.start()
        self.assertTrue(s.serving())
        server_addr = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        self.assertFalse(c.connected())
        c.connect(host=server_addr[0], port=server_addr[1])
        self.assertTrue(c.connected())
        time.sleep(WAIT_TIME)

        # Check addresses match
        self.assertEqual(s.get_addr(), c.get_server_addr())
        self.assertEqual(s.get_client_addr(0), c.get_addr())

        # Disconnect client
        c.disconnect()
        self.assertFalse(c.connected())
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        self.assertFalse(s.serving())
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_send_receive(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 1,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": 1,
            "client disconnected": 0
        })

        # Messages
        server_message = "Hello, server!"
        client_message = "Hello, client!"

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            self.assertEqual(data, server_message)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            self.assertEqual(data, client_message)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        c.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Send messages
        c.send(server_message)
        s.send(client_message)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_send_different_types(self):
        # Messages
        server_messages = [False, 11, "Hello from client #0!", 1.61803398, None]
        client_messages = ["Hello from the server!", 2.718, 2 ** 64, True]
        received_server_messages = []
        received_client_messages = []

        # Create expect map
        expected = ExpectMap({
            "server receive": len(server_messages),
            "server connect": 1,
            "server disconnect": 1,
            "client receive": len(client_messages),
            "client disconnected": 0
        })

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            received_server_messages.append(data)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            received_client_messages.append(data)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        c.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Send messages
        for server_message in server_messages:
            c.send(server_message)
        for client_message in client_messages:
            s.send(client_message)
        time.sleep(WAIT_TIME)

        # Check received messages match
        self.assertEqual(received_server_messages, server_messages)
        self.assertEqual(received_client_messages, client_messages)

        # Disconnect client
        c.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_send_large_messages(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 1,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": 1,
            "client disconnected": 0
        })

        # Messages
        large_server_message = os.urandom(random.randrange(32768, 65536))
        large_client_message = os.urandom(random.randrange(16384, 32768))

        # print(f"Large server message generated ({len(large_server_message)} bytes)")
        # print(f"Large client message generated ({len(large_client_message)} bytes)")

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            self.assertEqual(data, large_server_message)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            self.assertEqual(data, large_client_message)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        c.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Send messages
        c.send(large_server_message)
        s.send(large_client_message)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_sending_numerous_messages(self):
        # Messages
        num_server_messages = random.randrange(64, 128)
        num_client_messages = random.randrange(128, 256)
        server_messages = []
        client_messages = []
        received_server_messages = []
        received_client_messages = []
        for i in range(num_server_messages):
            server_messages.append(random.randrange(1024))
        for i in range(num_client_messages):
            client_messages.append(random.randrange(1024))
        # print(f"Generated {num_server_messages} server messages")
        # print(f"Generated {num_client_messages} client messages")

        # Create expect map
        expected = ExpectMap({
            "server receive": num_server_messages,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": num_client_messages,
            "client disconnected": 0
        })

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            received_server_messages.append(data)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            received_client_messages.append(data)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        c.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Send messages
        for server_message in server_messages:
            c.send(server_message)
        for client_message in client_messages:
            s.send(client_message)
        time.sleep(WAIT_TIME)

        # Check messages match
        self.assertEqual(server_messages, received_server_messages)
        self.assertEqual(client_messages, received_client_messages)

        # Disconnect client
        c.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_sending_custom_types(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 1,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": 1,
            "client disconnected": 0
        })

        # Messages
        server_message = Custom(123, "Hello, custom server class!", ["first server item", "second server item"])
        client_message = Custom(456, "Hello, custom client class!",
                                ["#1 client item", "client item #2", "(3) client item"])

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            self.assertEqual(data, server_message)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            self.assertEqual(data, client_message)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        c.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Send messages
        c.send(server_message)
        s.send(client_message)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_multiple_clients(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 2,
            "server connect": 2,
            "server disconnect": 2,
            "client 1 receive": 2,
            "client 2 receive": 2,
            "client disconnected": 0
        })

        # Messages
        message_from_client1 = "Hello from client #1!"
        message_from_client2 = "Goodbye from client #2!"
        message_from_server = "Hello from the server :)"
        receiving_message_from_server = False

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertTrue(client_id in [0, 1])

            if client_id == 0:
                self.assertEqual(data, message_from_client1)
            elif client_id == 1:
                self.assertEqual(data, message_from_client2)

            s.send(len(data), client_id)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertTrue(client_id in [0, 1])

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertTrue(client_id in [0, 1])

        def client1_on_receive(data):
            expected.received("client 1 receive")

            if receiving_message_from_server:
                self.assertEqual(data, message_from_server)
            else:
                self.assertEqual(data, len(message_from_client1))

        def client2_on_receive(data):
            expected.received("client 2 receive")

            if receiving_message_from_server:
                self.assertEqual(data, message_from_server)
            else:
                self.assertEqual(data, len(message_from_client2))

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client 1
        c1 = Client(on_receive=client1_on_receive, on_disconnected=client_on_disconnected)
        c1.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Check client 1 address info
        self.assertEqual(c1.get_addr(), s.get_client_addr(0))
        self.assertEqual(c1.get_server_addr(), s.get_addr())

        # Create client 2
        c2 = Client(on_receive=client2_on_receive, on_disconnected=client_on_disconnected)
        c2.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Check client 2 address info
        self.assertEqual(c2.get_addr(), s.get_client_addr(1))
        self.assertEqual(c2.get_server_addr(), s.get_addr())

        # Send message from client 1
        c1.send(message_from_client1)
        time.sleep(WAIT_TIME)

        # Send message from client 2
        c2.send(message_from_client2)
        time.sleep(WAIT_TIME)

        # Send message to all clients
        receiving_message_from_server = True
        s.send(message_from_server)
        time.sleep(WAIT_TIME)

        # Disconnect client 1
        c1.disconnect()
        time.sleep(WAIT_TIME)

        # Disconnect client 2
        c2.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_client_disconnected(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 0,
            "server connect": 1,
            "server disconnect": 0,
            "client receive": 0,
            "client disconnected": 1
        })

        def server_on_receive(client_id, data):
            expected.received("server receive")

        def server_on_connect(client_id):
            expected.received("server connect")

        def server_on_disconnect(client_id):
            expected.received("server disconnect")

        def client_on_receive(data):
            expected.received("client receive")

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        self.assertFalse(s.serving())
        s.start()
        self.assertTrue(s.serving())
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        self.assertFalse(c.connected())
        c.connect(host=server_host, port=server_port)
        self.assertTrue(c.connected())
        time.sleep(WAIT_TIME)

        # Stop server
        self.assertTrue(s.serving())
        self.assertTrue(c.connected())
        s.stop()
        self.assertFalse(s.serving())
        time.sleep(WAIT_TIME)
        self.assertFalse(c.connected())

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_remove_client(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 0,
            "server connect": 1,
            "server disconnect": 0,
            "client receive": 0,
            "client disconnected": 1
        })

        def server_on_receive(client_id, data):
            expected.received("server receive")

        def server_on_connect(client_id):
            expected.received("server connect")

        def server_on_disconnect(client_id):
            expected.received("server disconnect")

        def client_on_receive(data):
            expected.received("client receive")

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        s = Server(on_receive=server_on_receive, on_connect=server_on_connect, on_disconnect=server_on_disconnect)
        s.start()
        server_host, server_port = s.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c = Client(on_receive=client_on_receive, on_disconnected=client_on_disconnected)
        self.assertFalse(c.connected())
        c.connect(host=server_host, port=server_port)
        self.assertTrue(c.connected())
        time.sleep(WAIT_TIME)

        # Disconnect the client
        self.assertTrue(c.connected())
        s.remove_client(0)
        time.sleep(WAIT_TIME)
        self.assertFalse(c.connected())

        # Stop server
        s.stop()
        time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())

    def test_server_client_address_defaults(self):
        # Create server with host
        s1 = Server()
        s1.start(host="127.0.0.1")
        server_host, server_port = s1.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c1 = Client()
        c1.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c1.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s1.stop()
        time.sleep(WAIT_TIME)

        # Create server with port
        s2 = Server()
        s2.start(port=35792)
        server_host, server_port = s2.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c2 = Client()
        c2.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c2.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s2.stop()
        time.sleep(WAIT_TIME)

        # Create server with host and port
        s3 = Server()
        s3.start(host="127.0.0.1", port=35792)
        server_host, server_port = s3.get_addr()
        time.sleep(WAIT_TIME)

        # Create client
        c3 = Client()
        c3.connect(host=server_host, port=server_port)
        time.sleep(WAIT_TIME)

        # Disconnect client
        c3.disconnect()
        time.sleep(WAIT_TIME)

        # Stop server
        s3.stop()
        time.sleep(WAIT_TIME)

    def test_with_statements(self):
        # Create expect map
        expected = ExpectMap({
            "server receive": 1,
            "server connect": 1,
            "server disconnect": 1,
            "client receive": 1,
            "client disconnected": 0
        })

        # Messages
        server_message = "Hello, server!"
        client_message = "Hello, client #0!"

        def server_on_receive(client_id, data):
            expected.received("server receive")
            self.assertEqual(client_id, 0)
            self.assertEqual(data, server_message)

        def server_on_connect(client_id):
            expected.received("server connect")
            self.assertEqual(client_id, 0)

        def server_on_disconnect(client_id):
            expected.received("server disconnect")
            self.assertEqual(client_id, 0)

        def client_on_receive(data):
            expected.received("client receive")
            self.assertEqual(data, client_message)

        def client_on_disconnected():
            expected.received("client disconnected")

        # Create server
        with server(on_receive=server_on_receive,
                    on_connect=server_on_connect,
                    on_disconnect=server_on_disconnect) as s:
            server_host, server_port = s.get_addr()

            # Create client
            with client(host=server_host,
                        port=server_port,
                        on_receive=client_on_receive,
                        on_disconnected=client_on_disconnected) as c:
                c.send(server_message)
                s.send(client_message)
                time.sleep(WAIT_TIME)

            time.sleep(WAIT_TIME)

        # Check expect map
        self.assertEqual(expected.remaining(), {})
        self.assertTrue(expected.done())


if __name__ == "__main__":
    unittest.main()



import unittest

if __name__ == "__main__":
    test_suite = unittest.defaultTestLoader.discover("./test", pattern="test*.py")
    test_runner = unittest.TextTestRunner(verbosity=2)
    test_runner.run(test_suite)



from typing import Dict


class ExpectMap:
    def __init__(self, expected: Dict[str, int]) -> None:
        self._expected: Dict[str, int] = expected

    def received(self, name: str) -> None:
        self._expected[name] -= 1

    def get_expected(self) -> Dict[str, int]:
        return self._expected

    def remaining(self) -> Dict[str, int]:
        remaining = {}

        for name in self._expected:
            if self._expected[name] != 0:
                remaining[name] = self._expected[name]

        return remaining

    def done(self) -> bool:
        return len(self.remaining()) == 0



import os
from typing import Tuple

import rsa
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from rsa import PublicKey, PrivateKey

RSA_KEY_SIZE = 512
AES_KEY_SIZE = 32
AES_NONCE_SIZE = 12


def new_rsa_keys(key_size: int = RSA_KEY_SIZE) -> Tuple[PublicKey, PrivateKey]:
    """Generate a new RSA key pair."""

    public_key, private_key = rsa.newkeys(key_size)
    return public_key, private_key


def rsa_encrypt(public_key: PublicKey, plaintext: bytes) -> bytes:
    """Encrypt data using RSA."""

    ciphertext = rsa.encrypt(plaintext, public_key)
    return ciphertext


def rsa_decrypt(private_key: PrivateKey, ciphertext: bytes) -> bytes:
    """Decrypt data using RSA."""

    plaintext = rsa.decrypt(ciphertext, private_key)
    return plaintext


def new_aes_key(key_size: int = AES_KEY_SIZE) -> bytes:
    """Generate a new AES key."""

    key = os.urandom(key_size)
    return key


def aes_encrypt(key: bytes, plaintext: bytes) -> bytes:
    """Encrypt data using AES."""

    cipher = AESGCM(key)
    nonce = os.urandom(AES_NONCE_SIZE)
    ciphertext = cipher.encrypt(nonce, plaintext, b"")
    ciphertext_with_nonce = nonce + ciphertext
    return ciphertext_with_nonce


def aes_decrypt(key: bytes, ciphertext_with_nonce: bytes) -> bytes:
    """Decrypt data using AES."""

    cipher = AESGCM(key)
    nonce = ciphertext_with_nonce[:AES_NONCE_SIZE]
    ciphertext = ciphertext_with_nonce[AES_NONCE_SIZE:]
    plaintext = cipher.decrypt(nonce, ciphertext, b"")
    return plaintext



import errno
import pickle
import socket
import threading
import time
from contextlib import contextmanager
from typing import Any, Tuple, Union, Generator, Callable

import select

from .crypto import new_aes_key, rsa_encrypt
from .util import LEN_SIZE, DEFAULT_HOST, DEFAULT_PORT, LOCAL_SERVER_HOST, LOCAL_SERVER_PORT, encode_message_size, \
    decode_message_size, construct_message, deconstruct_message


class Client:
    """A socket client."""

    def __init__(self,
                 on_receive: Callable[[Any], None] = None,
                 on_disconnected: Callable[[], None] = None) -> None:
        """`on_receive` is a function that will be called when a message is received from the server.
        It takes one parameter: the data received.

        `on_disconnected` is a function that will be called when the client is unexpected disconnected from the server.
        It takes no parameters."""

        self._on_receive: Callable[[Any], None] = on_receive
        self._on_disconnected: Callable[[], None] = on_disconnected
        self._connected: bool = False
        self._key: Union[bytes, None] = None
        self._handle_thread: Union[threading.Thread, None] = None
        self._sock: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._local_server: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def connect(self, host: str = None, port: int = None) -> None:
        """Connect to a server."""

        if self._connected:
            raise RuntimeError("client is already connected to a server")

        if host is None:
            host = DEFAULT_HOST
        if port is None:
            port = DEFAULT_PORT

        self._sock.connect((host, port))
        self._connected = True
        self._exchange_keys()

        self._handle_thread = threading.Thread(target=self._handle)
        self._handle_thread.daemon = True
        self._handle_thread.start()

    def disconnect(self) -> None:
        """Disconnect from the server."""

        if not self._connected:
            raise RuntimeError("client is not connected to a server")

        self._connected = False

        local_client_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            local_client_sock.connect(self._local_server.getsockname())
        except ConnectionResetError:
            pass  # Connection reset by peer

        time.sleep(0.01)
        local_client_sock.close()
        self._local_server.close()

        self._sock.close()
        self._key = None

        if self._handle_thread is not None:
            if self._handle_thread is not threading.current_thread():
                self._handle_thread.join()
            self._handle_thread = None

    def send(self, data: Any) -> None:
        """Send data to the server."""

        if not self._connected:
            raise RuntimeError("client is not connected to a server")

        message = construct_message(data, self._key)
        self._sock.send(message)

    def connected(self) -> bool:
        """Check whether the client is connected to a server."""

        return self._connected

    def get_addr(self) -> Tuple[str, int]:
        """Get the address of the client."""

        if not self._connected:
            raise RuntimeError("client is not connected to a server")

        return self._sock.getsockname()

    def get_server_addr(self) -> Tuple[str, int]:
        """Get the address of the server."""

        if not self._connected:
            raise RuntimeError("client is not connected to a server")

        return self._sock.getpeername()

    def _handle(self) -> None:
        """Handle events from the server."""

        self._local_server.bind((LOCAL_SERVER_HOST, LOCAL_SERVER_PORT))
        self._local_server.listen()

        while self._connected:
            socks = [self._sock, self._local_server]
            select.select(socks, [], socks)

            if not self._connected:
                return

            try:
                size = self._sock.recv(LEN_SIZE)

                if len(size) == 0:
                    if self._connected:
                        self.disconnect()
                        self._call_on_disconnected()

                    return

                message_size = decode_message_size(size)
                message_encoded = self._sock.recv(message_size)
                message = deconstruct_message(message_encoded, self._key)

                self._call_on_receive(message)
            except ConnectionResetError:
                self.disconnect()
                self._call_on_disconnected()
                return
            except OSError as e:
                if e.errno == errno.ENOTSOCK:
                    self.disconnect()
                    self._call_on_disconnected()
                    return
                elif e.errno == errno.ECONNABORTED and not self._connected:
                    return
                elif e.errno == errno.EAGAIN or e.errno == errno.EWOULDBLOCK:
                    continue
                else:
                    raise e

    def _exchange_keys(self) -> None:
        """Exchange crypto keys with the server."""

        size_encoded = self._sock.recv(LEN_SIZE)
        size = decode_message_size(size_encoded)
        public_key_serialized = self._sock.recv(size)
        public_key = pickle.loads(public_key_serialized)

        key = new_aes_key()
        key_encrypted = rsa_encrypt(public_key, key)
        size_encoded = encode_message_size(len(key_encrypted))
        self._sock.send(size_encoded + key_encrypted)
        self._key = key

    def _call_on_receive(self, data: Any) -> None:
        """Call the receive callback."""

        if self._on_receive is not None:
            t = threading.Thread(target=self._on_receive, args=(data,))
            t.daemon = True
            t.start()

    def _call_on_disconnected(self) -> None:
        """Call the disconnected callback."""

        if self._on_disconnected is not None:
            t = threading.Thread(target=self._on_disconnected)
            t.daemon = True
            t.start()


@contextmanager
def client(host: str = None,
           port: int = None,
           on_receive: Callable[[Any], None] = None,
           on_disconnected: Callable[[], None] = None) -> Generator[Client, None, None]:
    """Use socket clients in a with statement."""

    c = Client(on_receive=on_receive,
               on_disconnected=on_disconnected)
    c.connect(host=host, port=port)
    yield c
    c.disconnect()



import errno
import pickle
import socket
import threading
import time
from contextlib import contextmanager
from typing import Any, Tuple, List, Dict, Union, Generator, Callable

import select

from .crypto import new_rsa_keys, rsa_decrypt
from .util import LEN_SIZE, DEFAULT_HOST, DEFAULT_PORT, encode_message_size, decode_message_size, construct_message, \
    deconstruct_message


class Server:
    """A socket server."""

    def __init__(self,
                 on_receive: Callable[[int, Any], None] = None,
                 on_connect: Callable[[int], None] = None,
                 on_disconnect: Callable[[int], None] = None) -> None:
        """`on_receive` is a function that will be called when a message is received from a client.
        It takes two parameters: client ID and data received.

        `on_connect` is a function that will be called when a client connects.
        It takes one parameter: client ID.

        `on_disconnect` is a function that will be called when a client disconnects.
        It takes one parameter: client ID."""

        self._on_receive: Callable[[int, Any], None] = on_receive
        self._on_connect: Callable[[int], None] = on_connect
        self._on_disconnect: Callable[[int], None] = on_disconnect
        self._serving: bool = False
        self._clients: Dict[int, socket.socket] = {}
        self._keys: Dict[int, bytes] = {}
        self._next_client_id = 0
        self._serve_thread: Union[threading.Thread, None] = None
        self._sock: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def start(self, host: str = None, port: int = None) -> None:
        """Start the server."""

        if self._serving:
            raise RuntimeError("server is already serving")

        if host is None:
            host = DEFAULT_HOST
        if port is None:
            port = DEFAULT_PORT

        self._sock.bind((host, port))
        self._sock.listen()
        self._serving = True

        self._serve_thread = threading.Thread(target=self._serve)
        self._serve_thread.daemon = True
        self._serve_thread.start()

    def stop(self) -> None:
        """Stop the server."""

        if not self._serving:
            raise RuntimeError("server is not serving")

        self._serving = False

        local_client_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            local_client_sock.connect(self._sock.getsockname())
        except ConnectionResetError:
            pass  # Connection reset by peer

        time.sleep(0.01)
        local_client_sock.close()

        for client_id in self._clients:
            client_sock = self._clients[client_id]
            client_sock.close()

        self._sock.close()
        self._clients = {}
        self._keys = {}

        if self._serve_thread is not None:
            if self._serve_thread is not threading.current_thread():
                self._serve_thread.join()
            self._serve_thread = None

    def send(self, data: Any, *client_ids: int) -> None:
        """Send data to clients. If no client IDs are specified, data will be sent to all clients."""

        if not self._serving:
            raise RuntimeError("server is not serving")

        if not client_ids:
            client_ids = self._clients.keys()

        for client_id in client_ids:
            if client_id in self._clients.keys():
                key = self._keys[client_id]
                conn = self._clients[client_id]
                message = construct_message(data, key)
                conn.send(message)
            else:
                raise RuntimeError(f"client {client_id} does not exist")

    def serving(self) -> bool:
        """Check whether the server is serving."""

        return self._serving

    def get_addr(self) -> Tuple[str, int]:
        """Get the address of the server."""

        if not self._serving:
            raise RuntimeError("server is not serving")

        return self._sock.getsockname()

    def get_client_addr(self, client_id: int) -> Tuple[str, int]:
        """Get the address of a client."""

        if not self._serving:
            raise RuntimeError("server is not serving")

        return self._clients[client_id].getpeername()

    def remove_client(self, client_id: int) -> None:
        """Remove a client."""

        if not self._serving:
            raise RuntimeError("server is not serving")

        if client_id not in self._clients.keys():
            raise RuntimeError(f"client {client_id} does not exist")

        conn = self._clients.pop(client_id)
        conn.shutdown(socket.SHUT_RDWR)
        conn.close()
        self._keys.pop(client_id)

    def _serve(self) -> None:
        """Serve clients."""

        while self._serving:
            try:
                socks: List[socket.socket] = list(self._clients.values())
                socks.insert(0, self._sock)
                select_result = select.select(socks, [], socks)
                read_socks: List[socket.socket] = select_result[0]
                exception_socks: List[socket.socket] = select_result[2]
            except ValueError:  # happens when a client is removed
                continue

            if not self._serving:
                return

            for notified_sock in read_socks:
                if notified_sock == self._sock:
                    try:
                        conn, _ = self._sock.accept()
                    except OSError as e:
                        if e.errno == errno.ENOTSOCK and not self._serving:
                            return
                        else:
                            raise e

                    client_id = self._new_client_id()

                    self._exchange_keys(client_id, conn)
                    self._clients[client_id] = conn
                    self._call_on_connect(client_id)
                else:
                    client_id = None

                    for sock_client_id in self._clients:
                        if self._clients[sock_client_id] == notified_sock:
                            client_id = sock_client_id

                    if client_id is not None:
                        try:
                            size = notified_sock.recv(LEN_SIZE)

                            if len(size) == 0:
                                try:
                                    self.remove_client(client_id)
                                except ValueError:
                                    pass

                                self._call_on_disconnect(client_id)
                                continue

                            message_size = decode_message_size(size)
                            message_encoded = notified_sock.recv(message_size)
                            key = self._keys[client_id]
                            message = deconstruct_message(message_encoded, key)

                            self._call_on_receive(client_id, message)
                        except OSError as e:
                            if e.errno == errno.ECONNRESET or e.errno == errno.ENOTSOCK:
                                if not self._serving:
                                    return

                                try:
                                    self.remove_client(client_id)
                                except ValueError:
                                    pass

                                self._call_on_disconnect(client_id)
                                continue
                            else:
                                raise e
                    else:
                        notified_sock.close()

            for notified_sock in exception_socks:
                client_id = None

                for sock_client_id in self._clients:
                    if self._clients[sock_client_id] == notified_sock:
                        client_id = sock_client_id

                if client_id is not None:
                    try:
                        self.remove_client(client_id)
                    except ValueError:
                        pass

                    self._call_on_disconnect(client_id)
                else:
                    notified_sock.close()

    def _exchange_keys(self, client_id: int, conn: socket.socket) -> None:
        """Exchange crypto keys with a client."""

        public_key, private_key = new_rsa_keys()
        public_key_serialized = pickle.dumps(public_key)
        size_encoded = encode_message_size(len(public_key_serialized))
        conn.send(size_encoded + public_key_serialized)

        size_encoded = conn.recv(LEN_SIZE)
        size = decode_message_size(size_encoded)
        key_encrypted = conn.recv(size)
        key = rsa_decrypt(private_key, key_encrypted)
        self._keys[client_id] = key

    def _new_client_id(self) -> int:
        """Generate a new client ID."""

        client_id = self._next_client_id
        self._next_client_id += 1
        return client_id

    def _call_on_receive(self, client_id: int, data: Any) -> None:
        """Call the receive callback."""

        if self._on_receive is not None:
            t = threading.Thread(target=self._on_receive, args=(client_id, data))
            t.daemon = True
            t.start()

    def _call_on_connect(self, client_id: int) -> None:
        """Call the connect callback."""

        if self._on_connect is not None:
            t = threading.Thread(target=self._on_connect, args=(client_id,))
            t.daemon = True
            t.start()

    def _call_on_disconnect(self, client_id: int) -> None:
        """Call the disconnect callback."""

        if self._on_disconnect is not None:
            t = threading.Thread(target=self._on_disconnect, args=(client_id,))
            t.daemon = True
            t.start()


@contextmanager
def server(host: str = None,
           port: int = None,
           on_receive: Callable[[int, Any], None] = None,
           on_connect: Callable[[int], None] = None,
           on_disconnect: Callable[[int], None] = None) -> Generator[Server, None, None]:
    """Use socket servers in a with statement."""

    s = Server(on_receive=on_receive,
               on_connect=on_connect,
               on_disconnect=on_disconnect)
    s.start(host=host, port=port)
    yield s
    s.stop()



import pickle
import socket
from typing import Any

from .crypto import aes_encrypt, aes_decrypt

LEN_SIZE: int = 5

DEFAULT_HOST: str = socket.gethostbyname(socket.gethostname())
DEFAULT_PORT: int = 29275

LOCAL_SERVER_HOST: str = "127.0.0.1"
LOCAL_SERVER_PORT: int = 0


def encode_message_size(size: int) -> bytes:
    """Encode the size portion of a message."""

    encoded_size = b""

    for i in range(LEN_SIZE):
        encoded_size = bytes([size & 0xff]) + encoded_size
        size >>= 8

    return encoded_size


def decode_message_size(encoded_size: bytes) -> int:
    """Decode the size portion of a message."""

    size = 0

    for i in range(LEN_SIZE):
        size <<= 8
        size += encoded_size[i]

    return size


def construct_message(data: Any, key: bytes) -> bytes:
    """Construct a message to be sent through a socket."""

    message_serialized = pickle.dumps(data)
    message_encrypted = aes_encrypt(key, message_serialized)
    message_size = encode_message_size(len(message_encrypted))
    return message_size + message_encrypted


def deconstruct_message(data: bytes, key: bytes) -> Any:
    """Deconstruct a message that came from a socket."""

    message_decrypted = aes_decrypt(key, data)
    message_deserialized = pickle.loads(message_decrypted)
    return message_deserialized
